from typing import Tuple, List

import pandas as pd
import numpy as np
from sklearn.preprocessing import OneHotEncoder, MinMaxScaler
from sklearn.compose import ColumnTransformer


def make_data_transformer(
    data: pd.DataFrame,
    numeric_columns: List[str],
    categorical_columns: List[str],
    numeric_transformer_cls=MinMaxScaler,
    categorical_transformer_cls=OneHotEncoder,
) -> ColumnTransformer:
    """Factory for ColumnTransformer"""
    transformer_list = []
    if numeric_transformer_cls != "passthrough":
        transformer = numeric_transformer_cls()
    else:
        transformer = "passthrough"
    transformer_list.append(("Numeric", transformer, numeric_columns))

    if categorical_transformer_cls == OneHotEncoder:
        transformer = OneHotEncoder(sparse=False, handle_unknown="ignore")
    else:
        transformer = categorical_transformer_cls(
            handle_unknown="use_encoded_value", unknown_value=-1
        )
    transformer_list.append(("Categorical", transformer, categorical_columns))
    column_transformer = ColumnTransformer(
        transformers=transformer_list, remainder="drop"
    )
    column_transformer.fit(data)
    return column_transformer


def filter_null_and_match_row_counts(real_data, synth_data, cols):
    num_null_values = real_data[cols].isna().sum().sum()
    if num_null_values > 0:
        real_data = real_data.dropna(subset=cols)
        synth_data = synth_data.head(len(real_data))
        print(
            str(num_null_values)
            + " null values found across columns "
            + ", ".join(cols)
            + ". Removing rows with null values and limiting synthesized row count to non-null real data row count."
        )

    return real_data, synth_data



from cgi import FieldStorage
from typing import Tuple, List, Dict

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from tonic_reporting.util import filter_null_and_match_row_counts
from scipy.stats import ks_2samp, anderson_ksamp, chisquare
from scipy.special import kl_div

from tonic_reporting.styling import Colors, PlotsSizes, Labels


def summarize_numeric(
    real_data: pd.DataFrame, synth_data: pd.DataFrame, columns: List[str]
) -> pd.DataFrame:
    rows = []
    for col in columns:
        real = real_data[col]
        fake = synth_data[col]
        anderson_stat, _, anderson_p_val = anderson_ksamp([real, fake])
        ks_stat, ks_p_val = ks_2samp(real, fake)
        row = {
            "Column": col,
            "Anderson-Darling statistic": anderson_stat,
            "Anderson-Darling p-value": anderson_p_val,
            "Kolmogorov-Smirnov statistic": ks_stat,
            "Kolmogorov-Smirnov p-value": ks_p_val,
        }

        rows.append(row)
    return pd.DataFrame(rows)


def get_frequencies(
    real_series: pd.Series, synthetic_series: pd.Series
) -> Tuple[np.ndarray, np.ndarray]:
    """Gets frequencies for each value in real and synthetic series. Assumes synthetic series values
    are subset of series_values
    """
    real_ft = real_series.value_counts()
    fake_ft = synthetic_series.value_counts()
    real_counts = []
    fake_counts = []
    for value in real_ft.keys():
        real_counts.append(real_ft[value])
        fake_counts.append(fake_ft.get(value, 0))
    return np.array(real_counts), np.array(fake_counts)


def chi_square_test(
    real_data: pd.Series, synthetic_data: pd.Series
) -> Tuple[float, float]:
    real_counts, fake_counts = get_frequencies(
        real_data,
        synthetic_data,
    )
    return chisquare(fake_counts, real_counts)


def kl_divergence(real_data: pd.Series, synthetic_data: pd.Series) -> float:
    """Computes D_KL(synthetic_data || real_data)"""
    real_counts, fake_counts = get_frequencies(
        real_data,
        synthetic_data,
    )
    idx = np.where(fake_counts != 0)[0]
    real_probs = real_counts[idx] / sum(real_counts[idx])
    fake_probs = fake_counts[idx] / sum(fake_counts[idx])

    return np.sum(fake_probs * np.log(fake_probs / real_probs))


def summarize_categorical(
    real_data: pd.DataFrame, synth_data: pd.DataFrame, columns: List[str]
) -> pd.DataFrame:
    rows = []
    for col in columns:
        real = real_data[col]
        fake = synth_data[col]

        cs_stat, cs_p_val = chi_square_test(real, fake)
        kl = kl_divergence(real_data[col], synth_data[col])
        row = {
            "Column": col,
            "Chi-Square statistic": cs_stat,
            "Chi-Square p-value": cs_p_val,
            "D_KL(Synth || Real)": kl,
        }
        rows.append(row)
    return pd.DataFrame(rows)


def plot_histogram(
    real_data: pd.DataFrame,
    synth_data: pd.DataFrame,
    col: str,
    ax=None,
    side_by_side=False,
    lower_quantile_lim=0.05,
    upper_quantile_lim=0.95,
):
    real_data, synth_data = filter_null_and_match_row_counts(
        real_data, synth_data, [col]
    )

    if ax is None:
        fig, ax = plt.subplots(1, 1)
    bins = np.histogram_bin_edges(np.array(real_data[col]), bins="auto")
    if side_by_side:
        ax.hist(
            [real_data[col], synth_data[col]],
            bins=bins,
            color=[Colors.REAL, Colors.SYNTH],
            label=[Labels.REAL, Labels.SYNTH],
        )
    else:
        ax.hist(
            real_data[col],
            bins=bins,
            color=Colors.REAL,
            alpha=Colors.ALPHA,
            label=Labels.REAL,
        )
        ax.hist(
            synth_data[col],
            bins=bins,
            color=Colors.SYNTH,
            alpha=Colors.ALPHA,
            label=Labels.SYNTH,
        )
    ax.set_xlim(
        real_data[col].quantile(lower_quantile_lim),
        real_data[col].quantile(upper_quantile_lim),
    )
    ax.legend()
    ax.set_title(col)
    ax.tick_params(axis="both", which="major")


def plot_frequency_table(
    real_data: pd.DataFrame,
    synth_data: pd.DataFrame,
    col: str,
    ax=None,
):

    vcs = pd.concat(
        [real_data[col].value_counts(), synth_data[col].value_counts()], axis=1
    )
    vcs.columns = ["Real", "Synthetic"]
    if vcs.shape[0] > 50:
        figsize = PlotsSizes.LARGE_FIGURE
    else:
        figsize = None
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    x_labels = list(vcs.index)
    x_axis = np.arange(len(x_labels))
    width = 0.2
    ax.bar(x_axis - width, vcs["Real"], 0.4, label="Real", color=Colors.REAL)
    ax.bar(x_axis + width, vcs["Synthetic"], 0.4, label="Synthetic", color=Colors.SYNTH)
    ax.set_xticks(x_axis)
    ax.set_xticklabels(x_labels, rotation="vertical")
    ax.legend()
    ax.set_title(col)



class Colors:
    REAL = "mediumpurple"
    SYNTH = "mediumturquoise"
    ALPHA = 0.5


class PlotsSizes:
    LARGE_FIGURE = (16, 10)
    XTRA_LARGE_FIGURE = (20, 10)


class Labels:
    REAL = "Real"
    SYNTH = "Synth"



from typing import Tuple, List

import pandas as pd
import numpy as np
from tonic_reporting.util import make_data_transformer, filter_null_and_match_row_counts
from sklearn.neighbors import BallTree
from sklearn.neighbors import NearestNeighbors


def compute_dcr(
    real_data: pd.DataFrame,
    synth_data: pd.DataFrame,
    numeric_columns: List[str],
    categorical_columns: List[str],
) -> Tuple[np.ndarray, np.ndarray]:
    real_data, synth_data = filter_null_and_match_row_counts(
        real_data, synth_data, numeric_columns + categorical_columns
    )

    """Computes Distances-to-Closest-Record for synth-to-real and real-to-real
    Real and synthetic data is transformed to unit square in R^n by transforming
    numeric columns with MinMaxScaler and categorical columns via OneHotEncoder.
    Nearest-neighbors are computed and the corresponding distances returned.
    """
    column_transformer = make_data_transformer(
        real_data, numeric_columns, categorical_columns
    )
    X_real = column_transformer.transform(real_data)
    X_synth = column_transformer.transform(synth_data)
    tree = BallTree(X_real, leaf_size=2)
    syn_distances, syn_indices = tree.query(X_synth, k=1)
    nbrs = NearestNeighbors(n_neighbors=2, algorithm="ball_tree").fit(X_real)
    real_distances, real_indices = nbrs.kneighbors(X_real)

    return syn_distances[:, 0], real_distances[:, 1]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tonic_reporting']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.0.0,<4.0.0',
 'numpy>=1.0.0,<2.0.0',
 'pandas>=1.0.0,<2.0.0',
 'scikit-learn>=1.0.0,<2.0.0',
 'scipy>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'tonic-reporting',
    'version': '1.0.0',
    'description': 'Tools for evaluating fidelity and privacy of synthetic data',
    'long_description': "# Overview\nThis library contains tools for evaluating fidelity and privacy of synthetic data.\n\n## Usage\n\nImport the desired modules from the library:\n\n```\nimport pandas as pd\nimport numpy as np\nimport matplotlib.pyplot as plt\nfrom tonic_reporting import univariate, multivariate, privacy\n```\n\n**Preface**\n\n*Numeric* columns refer to columns *encoded* as numeric. Numerical data types in the schema underlying a model may be encoded as other types.\n\n*Categorical* columns refer to columns *encoded* as categorical.\n\n*source_df* is a Pandas DataFrame of original data from the source database\n\n*synth_df* is a Pandas DataFrame of sampled data from trained models\n\nThe source and synthetic DataFrames should be equal in row count and schema.\n\n**Numeric Column Statistics**\n\n`univariate.summarize_numeric(source_df, synth_df, numeric_cols)`\n\n**Categorical Column Statistics**\n\n`univariate.summarize_categorical(source_df, synth_df, categorical_cols)`\n\n**Numeric Column Comparative Histograms**\n\n```\nfig, axarr = plt.subplots(1, len(numeric_cols), figsize = (9,12))\naxarr = axarr.ravel()\n\nfor col, ax in zip(numeric_cols, axarr):\n    univariate.plot_histogram(source_df, synth_df, col,ax)\n```\n\n**Numeric Column Multivariate Correlations Table**\n\n`multivariate.summarize_correlations(source_df, synth_df, numeric_cols)`\n\n**Numeric Column Multivariate Correlations Heat Map**\n\n```\nfig, axarr = plt.subplots(1, 2, figsize=(13, 8))\nmultivariate.plot_correlations(source_df, synth_df, numeric_cols, axarr=axarr, )\nfig.tight_layout()\n```\n\n**Distance to Closest Record Comparison**\n\n```\nsyn_dcr, real_dcr = privacy.compute_dcr(source_df, synth_df, numeric_cols, categorical_cols)\n\nfig, ax = plt.subplots(1,1,figsize=(8,6))\nax.hist(real_dcr,bins=300,label = 'Real vs. real', color='mediumpurple');\nax.hist(syn_dcr,bins=300,label='Synthetic vs. real', color='mediumturquoise');\nax.tick_params(axis='both', which='major', labelsize=14)\nax.set_title('Distances to closest record',fontsize=22)\nax.legend(fontsize=16);\n```\n",
    'author': 'Eric Timmerman',
    'author_email': 'eric@tonic.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.tonic.ai/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from tonic_reporting.util import filter_null_and_match_row_counts
from scipy.stats import pearsonr, spearmanr, kendalltau
from typing import Tuple, List

from tonic_reporting.styling import PlotsSizes


def summarize_correlations(
    real_data: pd.DataFrame,
    synth_data: pd.DataFrame,
    numeric_columns: List[str],
    corr_type="spearman",
) -> pd.DataFrame:
    real_data, synth_data = filter_null_and_match_row_counts(
        real_data, synth_data, numeric_columns
    )

    N = len(numeric_columns)
    corr_data = []
    if corr_type == "pearson":
        corr_fn = pearsonr
    elif corr_type == "spearman":
        corr_fn = spearmanr
    elif corr_type == "kendall":
        corr_fn = kendalltau
    else:
        raise ValueError("corr_type not recognized")

    for i in range(N):
        for j in range(i):
            x_name = numeric_columns[i]
            y_name = numeric_columns[j]
            real_stat, _ = corr_fn(real_data[x_name], real_data[y_name])
            synth_stat, _ = corr_fn(synth_data[x_name], synth_data[y_name])
            row = {
                "Column 1": x_name,
                "Column 2": y_name,
                "Real Correlation": real_stat,
                "Synthetic Correlation": synth_stat,
            }
            corr_data.append(row)

    return pd.DataFrame(corr_data)


def plot_correlations(
    real_data: pd.DataFrame,
    synth_data: pd.DataFrame,
    columns: List[str],
    axarr=None,
    corr_type="spearman",
    title_fontsize=18,
    label_fontsize=10,
):
    real_data, synth_data = filter_null_and_match_row_counts(
        real_data, synth_data, columns
    )

    real_corr = real_data[columns].corr(method=corr_type)
    synth_corr = synth_data[columns].corr()
    if axarr is None:
        fig, axarr = plt.subplots(1, 2, figsize=PlotsSizes.XTRA_LARGE_FIGURE)

    im0 = axarr[0].imshow(real_corr, cmap="Purples")
    axarr[0].set_xticks(np.arange(len(columns)))
    axarr[0].set_yticks(np.arange(len(columns)))
    axarr[0].set_xticklabels(columns, rotation="vertical")
    axarr[0].set_yticklabels(columns)
    axarr[0].set_title("Real Data Correlations", fontsize=title_fontsize)

    im1 = axarr[1].imshow(synth_corr, cmap="Purples")
    axarr[1].set_xticks(np.arange(len(columns)))
    axarr[1].set_yticks(np.arange(len(columns)))
    axarr[1].set_xticklabels(columns, rotation="vertical")
    axarr[1].set_yticklabels(columns)
    axarr[1].set_title("Synthetic Data Correlations", fontsize=title_fontsize)

    axarr[0].tick_params(axis="both", which="major", labelsize=label_fontsize)
    axarr[1].tick_params(axis="both", which="major", labelsize=label_fontsize)
    plt.colorbar(im0, fraction=0.046, pad=0.04, ax=axarr[0])
    plt.colorbar(im1, fraction=0.046, pad=0.04, ax=axarr[1])
